# The Keryx CLI - code so good, it'll make you want to slap yo momma!

import commands
import logging
import os

import keryx
from keryx import consts
from keryx.project import Project
from keryx.download import download, decompress

from prompt import prompt

class CLIKeryx(prompt):
    """Command line interface for Keryx"""
    def start(self):
        """Start the Keryx command line"""
        self.register_command('help', _('Display a list of commands'), self.help)
        self.register_command('open', 
                              _('Open a project - \'open <project>\''),
                              self.open_project,
                              1)
        self.register_command('create',
                              _('Create a new project - \'create <project>\''),
                              self.create_project,
                              1)

        # Say hi.
        print _('Welcome to the %s v%s command line!' % \
                (consts.appName, consts.appVersion))
        print _('Type \'help\' for a list of commands.')

        # Set the prompt.
        self.prompt_text = '%s> ' % consts.appName

        # Start prompting.
        self.start_prompt()

    ### Meat and potatoes ###

    def create_project(self, project_name):
        """Create a new project"""
        project_name = project_name[0]  # Args come in a list

        # Append a new project
        keryx.project.projects.append(Project())
        self.project = keryx.project.projects[-1]

        # Load the plugins
        keryx.plugins.load(0, False)

        # Grab the plugin entry
        self.plugin = False
        os = 'Debian'
        for plugin in keryx.plugins.OSPluginList:
            if plugin[0] == os: 
                self.plugin = plugin
        if not self.plugin:
            logging.error(_('%s plugin not found') % os)
            return

        # Create project
        success, filename = self.project.CreateKeryx(project_name, 
                                                     self.plugin[0],
                                                     self.plugin[1])
        if not success:
            logging.error(_('Unable to create project. Make sure a project by ' \
                        'this name does not already exist.'))

    def open_project(self, project_name):
        """Open a project"""
        project_name = project_name[0]  # Args come in a list

        # Get the project's filename
        projects = get_projects()  # Get available project {names: filenames}
        if project_name in projects.keys():
            project_path = projects[project_name]
        elif project_name.endswith(consts.appFileExt) and \
             os.path.exists(project_name):
            # Userboy apparently gave the entire path to the project file.
            project_path = project_name
        else:
            logging.error(_('Unable to find project %s' % project_name))
            return

        # Load the plugins
        keryx.plugins.load(0, False)

        # Append a new project
        keryx.project.projects.append(Project())
        self.project = keryx.project.projects[-1]

        # Open the project from file
        success = self.project.OpenKeryx(project_path)
        if success:
            logging.info(_('Opened project %s' % self.project.name))
        else:
            # Remove project
            keryx.project.projects.pop()
            logging.error(_('Failed to open project'))
            return

        # Set new prompt text.
        self.prompt_text = '%s:%s> ' % (consts.appName, self.project.name)

        # Load package data.
        self.refresh() 

        # Register new commands available after loading project.
        self.register_command('close',
                              'Close the current project',
                              self.close,
                              0)
        self.register_command('download',
                              'Download listed packages and their dependencies',
                              self.download,
                              -1)
        self.register_command('install',
                              'Install listed packages',
                              self.install,
                              -1)
        self.register_command('refresh',
                              'Update package data',
                              self.refresh,
                              0)
        self.register_command('search',
                              'Search package names',
                              self.search,
                              -1)
        self.register_command('update',
                              'Download updates for out-of-date packages',
                              self.update,
                              0)
        self.register_command('updatestatus',
                              'Update the list of installed packages',
                              self.updatestatus,
                              0)

    def close(self):
        """Close the current project"""
        name = self.project.name  # It'll come in handy for logging.

        # Delete the project data.
        del(self.project)

        # Set prompt text.
        self.prompt_text = '%s> ' % consts.appName

        # Remove commands only available with open project.
        self.remove_command('close')
        self.remove_command('download')
        self.remove_command('install')
        self.remove_command('refresh')
        self.remove_command('search')
        self.remove_command('update')
        self.remove_command('updatestatus')

        logging.info(_('Closed project %s' % name))

    def download(self, packages):
        """Download the listed packages and their dependencies"""
        # Get the dependencies.
        files = []
        for package in packages:
            if package in self.project.packages.keys():
                dependencies = self.project.getDependencies(package)
                # If the files are not already to be downloaded, add them.
                files += [x for x in dependencies if not x in files]
            else:
                print _('Could not find package: %s' % package)

        # Make sure the destination dir exists.
        pkgdir = os.path.join(self.project.dir, 'packages')
        if not os.path.exists(pkgdir):
            os.mkdir(pkgdir)

        # Download the files.
        if files:  # If there's something to download.
            print _('The following packages will be downloaded:')
            # Print an indented list of files to be downloaded.
            print '\n'.join(['  %s' % x[0] for x in files])
            go_ahead = ask(_('Do you want to continue?'))

            if go_ahead:
                for afile in files:
                    # afile contains (source, destination, {hashtype: hash})
                    download(afile[0], afile[1], consts.http_proxy)
            else:
                return

    def install(self, packages):
        """Installs the listed packages"""
        # Update the APT cache.
        result = self.__install_cache()
        if not result:
            logging.error(_('Failed to update the APT cache. Aborting installation.'))
            return False

        # Install the packages.
        run = self.project.plugin.installPacksCommand(self.project.dir, 
                                                      ' '.join(packages))
        status, output = run_root(run)  # Run, root, run!
        if status == 0:
            logging.info(_('Successfully installed packages'))
        else:
            logging.error(_('Failed to install packages:\n%s' % output))

    def refresh(self):
        """Update package data"""
        update = ask(_('Would you like to download the latest package lists?'))
        if update:
            urls = self.project.getUrls()

            if urls:  # If there's something to download.
                print _('The following files will be downloaded:')
                # Print an indented list of files to be downloaded.
                print '\n'.join(['  %s' % x[0] for x in urls])
                go_ahead = ask(_('Do you want to continue?'))

                if go_ahead:
                    for afile in urls:
                        # afile contains (source, destination)
                        download(afile[0], afile[1], consts.http_proxy)
                        decompress(afile[1])

        # Load the package data.
        self.project.loadLocal(0, 0)
        # (0, 0) because loadLocal doesn't even use these args.  :-/

    def search(self, string):
        """Print package names, descriptions containing search string"""
        string = ' '.join(string)  # Put parts of search string together.
        #TODO: make sure descriptions are tabbed past the longest package names
        for package in sorted(self.project.packages.keys()):
            description = self.project.packages[package][4]
            if string in package or string in description:
                print '%s\t%s' % (package, description)

    def update(self):
        """Download updates for out-of-date packages"""
        # Get a list of out-of-date packages.
        updates = [x for (x, y) in self.project.packages.iteritems() if y[0] == 1]
        
        if updates:
            # Send the list off to be downloaded.
            self.download(updates)
        else:
            print _('All packages are up-to-date')

    def updatestatus(self):
        """Update the list of installed packages"""
        go_ahead = ask(_('You should only update the project\'s status on ' \
                            'the computer the project was created on. Do you ' \
                            'want to continue?'))
        if go_ahead:
            run = self.project.plugin.updateStatusCommand(self.project.dir)
            success = run_root(run)
            if not success:
                print logging.error(_('Failed to update status'))

    def __install_cache(self):
        """Updates the APT cache"""
        # Write the installCache script.
        result = self.project.plugin.installCache(self.project.dir, 'installCache')
        if not result:
            logging.error(_('Failed to update the APT cache.'))
            return False

        # Run the installCache script.
        run = self.project.plugin.installCacheCommand(self.project.dir, 'installCache')
        status, output = run_root(run)
        if status != 0:
            logging.error(_('installCache script failed with status %s: %s' % \
                        (status, output)))
            return False
        return True


def get_projects():
    """Return a dictionary of avalable projects' {names: filenames, ...}"""
    projects = {}
    for root, dirs, files in os.walk(consts.dirProjects):
        for item in files:
            if item.endswith(consts.appFileExt): 
                project_name = item[:-len(consts.appFileExt)]
                project_path = os.path.join(root, item)
                projects.update({project_name: project_path})
    return projects


def ask(question, default=True):
    """Asks the user a yes/no question"""
    # Keep asking until a valid answer is given.
    while True:
        if default:
            hint = '[Y/n]'
        else:
            hint = '[n/Y]'
        answer = raw_input(_('%s %s ') % (question, hint)).lower().strip()

        if not answer:
            # No answer. Go with default. Purdy safe bet.
            return default
        elif answer.startswith('y'):
            return True
        elif answer.startswith('n'):
            return False


def run_root(command):
    """Runs a command as root"""
    root = 'sudo'
    status, output = commands.getstatusoutput('%s %s' % (root, command))
    return status, output


def start():
    """Start the CLI"""
    keryx = cliKeryx()
    keryx.start()
